// ASEPRITE gui library
// Copyright (C) 2001-2012  David Capello
//
// This source file is ditributed under a BSD-like license, please
// read LICENSE.txt for more information.

#ifndef GUI_WIDGET_H_INCLUDED
#define GUI_WIDGET_H_INCLUDED

#include <string>

#include "gfx/border.h"
#include "gfx/rect.h"
#include "gfx/size.h"
#include "gui/base.h"
#include "gui/component.h"
#include "gui/rect.h"
#include "gui/widgets_list.h"

class PaintEvent;
class PreferredSizeEvent;
namespace gui { class Manager; }

#ifndef NDEBUG
#include "gui/intern.h"
#define ASSERT_VALID_WIDGET(widget) ASSERT((widget) != NULL &&          \
                                           _ji_is_valid_widget((widget)))
#else
#define ASSERT_VALID_WIDGET(widget) ((void)0)
#endif

struct FONT;
struct BITMAP;

class Frame;

typedef std::vector<Widget*> WidgetsList;

int ji_register_widget_type();

/* hooks */

void jwidget_add_hook(JWidget widget, int type,
                      MessageFunc msg_proc, void *data);
JHook jwidget_get_hook(JWidget widget, int type);
void *jwidget_get_data(JWidget widget, int type);

/* position and geometry */

JRect jwidget_get_rect(JWidget widget);
JRect jwidget_get_child_rect(JWidget widget);
JRegion jwidget_get_region(JWidget widget);
JRegion jwidget_get_drawable_region(JWidget widget, int flags);
int jwidget_get_bg_color(JWidget widget);
int jwidget_get_text_length(JWidget widget);
int jwidget_get_text_height(JWidget widget);
void jwidget_get_texticon_info(JWidget widget,
                               JRect box, JRect text, JRect icon,
                               int icon_align, int icon_w, int icon_h);

void jwidget_noborders(JWidget widget);
void jwidget_set_border(JWidget widget, int value);
void jwidget_set_border(JWidget widget, int l, int t, int r, int b);
void jwidget_set_rect(JWidget widget, JRect rect);
void jwidget_set_min_size(JWidget widget, int w, int h);
void jwidget_set_max_size(JWidget widget, int w, int h);
void jwidget_set_bg_color(JWidget widget, int color);

/* signal handle */

void jwidget_signal_on(JWidget widget);
void jwidget_signal_off(JWidget widget);

bool jwidget_emit_signal(JWidget widget, int signal_num);

//////////////////////////////////////////////////////////////////////

class Widget : public Component
{
public:
  JID id;                       /* identify code */
  int type;                     /* widget's type */

  char *name;                   /* widget's name */
  JRect rc;                     /* position rectangle */
  struct {
    int l, t, r, b;
  } border_width;               /* border separation with the parent */
  int child_spacing;            /* separation between children */

  /* flags */
  int flags;
  int emit_signals;             /* emit signal counter */

  /* widget size limits */
  int min_w, min_h;
  int max_w, max_h;

  /* structures */
  JList children;                /* sub-objects */
  JWidget parent;                /* who is the parent? */

  /* virtual properties */
  JList hooks;                  /* hooks with msg_proc and specific data */

private:
  Theme* m_theme;               // Widget's theme
  int m_align;                  // Widget alignment
  std::string m_text;           // Widget text
  struct FONT *m_font;          // Text font type
  int m_bg_color;               // Background color
  JRegion m_update_region;      // Region to be redrawed.

public:
  // Extra data for the theme
  void *theme_data[4];

  /* for user */
  void *user_data[4];

  // ===============================================================
  // CTOR & DTOR
  // ===============================================================

  Widget(int type);
  virtual ~Widget();

  // Safe way to delete a widget when it is not in the manager message
  // queue anymore.
  void deferDelete();

  // main properties

  int getType() const;
  const char* getName() const;
  int getAlign() const;

  void setName(const char* name);
  void setAlign(int align);

  // text property

  bool hasText() const { return flags & JI_NOTEXT ? false: true; }

  const char* getText() const { return m_text.c_str(); }
  int getTextInt() const;
  double getTextDouble() const;
  size_t getTextSize() const { return m_text.size(); }
  void setText(const char* text);
  void setTextf(const char* text, ...);
  void setTextQuiet(const char* text);

  // ===============================================================
  // COMMON PROPERTIES
  // ===============================================================

  // True if this widget and all its ancestors are visible.
  bool isVisible() const;
  void setVisible(bool state);

  // True if this widget can receive user input (is not disabled).
  bool isEnabled() const;
  void setEnabled(bool state);

  // True if this widget is selected (pushed in case of a button, or
  // checked in the case of a check-box).
  bool isSelected() const;
  void setSelected(bool state);

  // True if this widget wants more space when it's inside a Box
  // parent.
  bool isExpansive() const;
  void setExpansive(bool state);

  // True if this is a decorative widget created by the current
  // theme. Decorative widgets are arranged by the theme instead that
  // the parent's widget.
  bool isDecorative() const;
  void setDecorative(bool state);

  // True if this widget can receive the keyboard focus.
  bool isFocusStop() const;
  void setFocusStop(bool state);

  // True if this widget wants the focus by default when it's shown by
  // first time (e.g. when its parent window is opened).
  void setFocusMagnet(bool state);
  bool isFocusMagnet() const;

  // ===============================================================
  // LOOK & FEEL
  // ===============================================================

  FONT* getFont();
  void setFont(FONT* font);

  // Gets the background color of the widget.
  int getBgColor() const {
    if (m_bg_color < 0 && parent)
      return parent->getBgColor();
    else
      return m_bg_color;
  }

  // Sets the background color of the widget
  void setBgColor(int bg_color) {
    m_bg_color = bg_color;
  }

  Theme* getTheme() const { return m_theme; }
  void setTheme(Theme* theme);

  void initTheme();

  // ===============================================================
  // PARENTS & CHILDREN
  // ===============================================================

  Frame* getRoot();
  Widget* getParent();
  gui::Manager* getManager();

  // Returns a list of parents (you must free the list), if
  // "ascendant" is true the list is build from child to parents, else
  // the list is from parent to children.
  JList getParents(bool ascendant);

  // Returns a list of children (you must free the list).
  JList getChildren();

  // Returns the next or previous siblings.
  Widget* getNextSibling();
  Widget* getPreviousSibling();

  Widget* pick(int x, int y);
  bool hasChild(Widget* child);
  Widget* findChild(const char* name);

  // Returns a widget in the same window that is located "sibling".
  Widget* findSibling(const char* name);

  // Finds a child with the specified name and dynamic-casts it to
  // type T.
  template<class T>
  T* findChildT(const char* name) {
    return dynamic_cast<T*>(findChild(name));
  }

  void addChild(Widget* child);
  void removeChild(Widget* child);
  void replaceChild(Widget* oldChild, Widget* newChild);
  void insertChild(int index, Widget* child);

  // ===============================================================
  // LAYOUT & CONSTRAINT
  // ===============================================================

  void layout();

  // ===============================================================
  // POSITION & GEOMETRY
  // ===============================================================

  gfx::Rect getBounds() const;
  gfx::Rect getClientBounds() const;
  void setBounds(const gfx::Rect& rc);

  gfx::Border getBorder() const;
  void setBorder(const gfx::Border& border);

  // ===============================================================
  // REFRESH ISSUES
  // ===============================================================

  bool isDoubleBuffered();
  void setDoubleBuffered(bool doubleBuffered);

  void invalidate();
  void invalidateRect(const gfx::Rect& rect);
  void invalidateRect(const JRect rect);
  void invalidateRegion(const JRegion region);

  void flushRedraw();

  void scrollRegion(JRegion region, int dx, int dy);

  // ===============================================================
  // GUI MANAGER
  // ===============================================================

  bool sendMessage(Message* msg);
  void closeWindow();

  void broadcastMouseMessage(WidgetsList& targets);

  // ===============================================================
  // SIZE & POSITION
  // ===============================================================

  gfx::Size getPreferredSize();
  gfx::Size getPreferredSize(const gfx::Size& fitIn);
  void setPreferredSize(const gfx::Size& fixedSize);
  void setPreferredSize(int fixedWidth, int fixedHeight);

  // ===============================================================
  // MOUSE, FOCUS & KEYBOARD
  // ===============================================================

  void requestFocus();
  void releaseFocus();
  void captureMouse();
  void releaseMouse();
  bool hasFocus();
  bool hasMouse();
  bool hasMouseOver();
  bool hasCapture();

  // Returns lower-case letter that represet the mnemonic of the widget
  // (the underscored character, i.e. the letter after & symbol).
  int getMnemonicChar() const;

  bool isScancodeMnemonic(int scancode) const;

protected:

  // ===============================================================
  // MESSAGE PROCESSING
  // ===============================================================

  virtual bool onProcessMessage(Message* msg);

  // ===============================================================
  // EVENTS
  // ===============================================================

  virtual void onInvalidateRegion(const JRegion region);
  virtual void onPreferredSize(PreferredSizeEvent& ev);
  virtual void onPaint(PaintEvent& ev);
  virtual void onBroadcastMouseMessage(WidgetsList& targets);

private:
  gfx::Size* m_preferredSize;
  bool m_doubleBuffered : 1;
};

#endif
